"""LISY platform for System 1 and System 80."""
import asyncio
from distutils.version import StrictVersion

from typing import Dict, Optional, List

from mpf.core.segment_mappings import SEVEN_SEGMENTS, BCD_SEGMENTS, FOURTEEN_SEGMENTS, TextToSegmentMapper, \
    ASCII_SEGMENTS
from mpf.core.platform_batch_light_system import PlatformBatchLight, PlatformBatchLightSystem
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings
from mpf.platforms.interfaces.hardware_sound_platform_interface import HardwareSoundPlatformInterface
from mpf.platforms.interfaces.segment_display_platform_interface import SegmentDisplaySoftwareFlashPlatformInterface

from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface

from mpf.core.logging import LogMixin
from mpf.core.utility_functions import Util

from mpf.platforms.lisy.defines import LisyDefines

from mpf.platforms.interfaces.light_platform_interface import LightPlatformSoftwareFade, LightPlatformInterface

from mpf.core.platform import SwitchPlatform, LightsPlatform, DriverPlatform, SwitchSettings, DriverSettings, \
    DriverConfig, SwitchConfig, SegmentDisplaySoftwareFlashPlatform, HardwareSoundPlatform, RepulseSettings


class LisySwitch(SwitchPlatformInterface):

    """A switch in the LISY platform."""

    __slots__ = ["index"]  # type: List[str]

    def __init__(self, config, number):
        """Initialise switch."""
        super().__init__(config, number)
        self.index = int(number)

    def get_board_name(self):
        """Return board name."""
        return "LISY"


class LisyDriver(DriverPlatformInterface):

    """A driver in the LISY platform."""

    __slots__ = ["platform", "_pulse_ms", "_recycle_time", "index", "has_rule"]     # type: List[str]

    def __init__(self, config, number, platform):
        """Initialise driver."""
        super().__init__(config, number)
        self.platform = platform
        self._pulse_ms = -1
        self._recycle_time = None
        self.index = int(number)
        self.has_rule = False

    def configure_recycle(self, recycle_time):
        """Configure recycle time."""
        if self.platform.api_version < StrictVersion("0.9"):
            return
        if recycle_time > 255:
            recycle_time = 255
        elif recycle_time < 0:
            recycle_time = 0

        if self._recycle_time != recycle_time:
            self._recycle_time = recycle_time
            self.platform.send_byte(LisyDefines.SetSolenoidsRecycleTime,
                                    bytes([int(self.number), recycle_time]))

    def _configure_pulse_ms(self, pulse_ms):
        """Configure pulse ms for this driver if it changed."""
        if pulse_ms != self._pulse_ms:
            self._pulse_ms = pulse_ms
            self.platform.send_byte(LisyDefines.SolenoidsSetSolenoidPulseTime, bytes(
                [int(self.number),
                 pulse_ms if pulse_ms < 255 else 255
                 ]))

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse driver."""
        self._configure_pulse_ms(pulse_settings.duration)
        self.platform.send_byte(LisyDefines.SolenoidsPulseSolenioid, bytes([int(self.number)]))

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable driver."""
        del hold_settings
        self._configure_pulse_ms(pulse_settings.duration)
        self.platform.send_byte(LisyDefines.SolenoidsSetSolenoidToOn, bytes([int(self.number)]))

    def disable(self):
        """Disable driver."""
        self.platform.send_byte(LisyDefines.SolenoidsSetSolenoidToOff, bytes([int(self.number)]))

    def get_board_name(self):
        """Return board name."""
        return "LISY"


class LisySimpleLamp(LightPlatformSoftwareFade):

    """A simple light in the LISY platform which only supports on/off."""

    __slots__ = ["platform", "_state"]

    def __init__(self, number, platform):
        """Initialise Lisy Light."""
        super().__init__(number, platform.machine.clock.loop, 50)
        self.platform = platform
        self._state = None

    def set_brightness(self, brightness: float):
        """Turn lamp on or off."""
        if brightness > 0 and self._state is not True:
            self.platform.send_byte(LisyDefines.LampsSetLampOn, bytes([self.number]))
            self._state = True
        elif brightness <= 0 and self._state is not False:
            self.platform.send_byte(LisyDefines.LampsSetLampOff, bytes([self.number]))
            self._state = False

    def get_board_name(self):
        """Return board name."""
        return "LISY"

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        return self.number == other.number + 1

    def get_successor_number(self):
        """Return next number."""
        return self.number + 1

    def __lt__(self, other):
        """Order lights by their order on the hardware."""
        return self.number < other.number


class LisyModernLight(PlatformBatchLight):

    """A modern light in LISY."""

    __slots__ = ["platform"]

    def __init__(self, number, platform, light_system):
        """Initialise Lisy Light."""
        super().__init__(number, light_system)
        self.platform = platform

    def get_max_fade_ms(self) -> int:
        """Return max fade time."""
        return 65535

    def get_board_name(self):
        """Return board name."""
        return "LISY"

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        return self.number == other.number + 1

    def get_successor_number(self):
        """Return next number."""
        return self.number + 1

    def __lt__(self, other):
        """Order lights by their order on the hardware."""
        return self.number < other.number


class LisyDisplay(SegmentDisplaySoftwareFlashPlatformInterface):

    """A segment display in the LISY platform."""

    __slots__ = ["platform", "_type_of_display", "_length_of_display"]

    def __init__(self, number: int, platform: "LisyHardwarePlatform") -> None:
        """Initialise segment display."""
        super().__init__(number)
        self.platform = platform
        self._type_of_display = None
        self._length_of_display = None

    async def initialize(self):
        """Initialize segment display."""
        if self.platform.api_version >= StrictVersion("0.9"):
            # display info for display
            display_info = await self.platform.send_byte_and_read_response(
                LisyDefines.InfoGetDisplayDetails, bytearray([self.number]), 2)
            if 1 > display_info[0] > 6:
                raise AssertionError("Invalid display type {} reported by hardware for display {}".format(
                    self._type_of_display, self.number))

            self._type_of_display = display_info[0]
            self._length_of_display = int(display_info[1])

        # clear display initially
        self._set_text("")

    def _format_text(self, text):
        if self._type_of_display == 1:
            mapping = TextToSegmentMapper.map_text_to_segments(text, self._length_of_display, BCD_SEGMENTS,
                                                               embed_dots=False)
            result = map(lambda x: x.get_x4x3x2x1_encoding(), mapping)
        elif self._type_of_display == 2:
            mapping = TextToSegmentMapper.map_text_to_segments(text, self._length_of_display, BCD_SEGMENTS)
            result = map(lambda x: x.get_dpx4x3x2x1_encoding(), mapping)
        elif self._type_of_display == 3:
            mapping = TextToSegmentMapper.map_text_to_segments(text, self._length_of_display, SEVEN_SEGMENTS)
            result = map(lambda x: x.get_dpgfeabcd_encoding(), mapping)
        elif self._type_of_display == 4:
            mapping = TextToSegmentMapper.map_text_to_segments(text, self._length_of_display, FOURTEEN_SEGMENTS)
            result = map(lambda x: x.get_apc_encoding(), mapping)
        elif self._type_of_display == 5:
            mapping = TextToSegmentMapper.map_text_to_segments(text, self._length_of_display, ASCII_SEGMENTS,
                                                               embed_dots=False)
            result = map(lambda x: x.get_ascii_encoding(), mapping)
        elif self._type_of_display == 6:
            mapping = TextToSegmentMapper.map_text_to_segments(text, self._length_of_display, ASCII_SEGMENTS)
            result = map(lambda x: x.get_ascii_with_dp_encoding(), mapping)
        else:
            raise AssertionError("Invalid type {}".format(self._type_of_display))

        return b''.join(result)

    def _set_text(self, text: str):
        """Set text to display."""
        assert self.platform.api_version is not None
        if self.platform.api_version >= StrictVersion("0.9"):
            formatted_text = self._format_text(text)
            self.platform.send_byte(LisyDefines.DisplaysSetDisplay0To + self.number,
                                    bytearray([len(formatted_text)]) + formatted_text)
        else:
            self.platform.send_string(LisyDefines.DisplaysSetDisplay0To + self.number, text)


class LisySound(HardwareSoundPlatformInterface):

    """Hardware sound interface for LISY."""

    __slots__ = ["platform"]

    def __init__(self, platform):
        """Initialise hardware sound."""
        self.platform = platform        # type: LisyHardwarePlatform

    def play_sound(self, number: int, track: int = 1):
        """Play sound with number."""
        assert self.platform.api_version is not None
        if self.platform.api_version >= StrictVersion("0.9"):
            self.platform.send_byte(LisyDefines.SoundPlaySound, bytes([track, number]))
        else:
            assert track == 1
            self.platform.send_byte(LisyDefines.SoundPlaySound, bytes([number]))

    def play_sound_file(self, file: str, platform_options: dict, track: int = 1):
        """Play sound file."""
        assert self.platform.api_version is not None
        flags = 1 if platform_options.get("loop", False) else 0
        flags += 2 if platform_options.get("no_cache", False) else 0
        if self.platform.api_version >= StrictVersion("0.9"):
            self.platform.send_string(LisyDefines.SoundPlaySoundFile, chr(track) + chr(flags) + file)
        else:
            assert track == 1
            self.platform.send_string(LisyDefines.SoundPlaySoundFile, chr(flags) + file)

    def text_to_speech(self, text: str, platform_options: dict, track: int = 1):
        """Text to speech."""
        assert self.platform.api_version is not None
        flags = 1 if platform_options.get("loop", False) else 0
        flags += 2 if platform_options.get("no_cache", False) else 0
        if self.platform.api_version >= StrictVersion("0.9"):
            self.platform.send_string(LisyDefines.SoundTextToSpeech, chr(track) + chr(flags) + text)
        else:
            assert track == 1
            self.platform.send_string(LisyDefines.SoundTextToSpeech, chr(flags) + text)

    def set_volume(self, volume: float, track: int = 1):
        """Set volume."""
        assert self.platform.api_version is not None
        if self.platform.api_version >= StrictVersion("0.9"):
            self.platform.send_byte(LisyDefines.SoundSetVolume, bytes([track, int(volume * 100)]))
        else:
            assert track == 1
            self.platform.send_byte(LisyDefines.SoundSetVolume, bytes([int(volume * 100)]))

    def stop_all_sounds(self, track: int = 1):
        """Stop all sounds."""
        assert self.platform.api_version is not None
        if self.platform.api_version >= StrictVersion("0.9"):
            self.platform.send_byte(LisyDefines.SoundStopAllSounds, bytes([track]))
        else:
            assert track == 1
            self.platform.send_byte(LisyDefines.SoundStopAllSounds)


# pylint: disable-msg=too-many-instance-attributes
class LisyHardwarePlatform(SwitchPlatform, LightsPlatform, DriverPlatform,
                           SegmentDisplaySoftwareFlashPlatform,
                           HardwareSoundPlatform, LogMixin):

    """LISY platform."""

    __slots__ = ["config", "_writer", "_reader", "_poll_task", "_watchdog_task", "_number_of_lamps",
                 "_number_of_solenoids", "_number_of_displays", "_inputs", "_coils_start_at_one",
                 "_bus_lock", "api_version", "_number_of_switches", "_number_of_modern_lights",
                 "_light_system", "_send_length_of_command", "_lisy_version", "_hardware_name"]  # type: List[str]

    def __init__(self, machine) -> None:
        """Initialise platform."""
        super().__init__(machine)
        self._writer = None                 # type: Optional[asyncio.StreamWriter]
        self._reader = None                 # type: Optional[asyncio.StreamReader]
        self._poll_task = None
        self._watchdog_task = None
        self._bus_lock = asyncio.Lock()
        self._number_of_lamps = None        # type: Optional[int]
        self._number_of_solenoids = None    # type: Optional[int]
        self._number_of_displays = None     # type: Optional[int]
        self._number_of_switches = None     # type: Optional[int]
        self._number_of_modern_lights = None    # type: Optional[int]
        self._inputs = dict()               # type: Dict[str, bool]
        self._coils_start_at_one = None     # type: Optional[str]
        self.features['max_pulse'] = 255
        self._lisy_version = None
        self._hardware_name = None

        self.config = self.machine.config_validator.validate_config("lisy", self.machine.config['lisy'])
        self._configure_device_logging_and_debug("lisy", self.config)
        self.api_version = None             # type: Optional[StrictVersion]
        self._light_system = None
        self._send_length_of_command = self.config['send_length_after_command']

    def _disable_dts_on_start_of_serial(self):
        """Prevent DTS toggling when opening the serial.

        This works only on unix. On other platforms import of termios will fail.
        """
        try:
            # pylint: disable-msg=import-outside-toplevel
            import termios
        except ImportError:
            self.warning_log("Could not import terminos (this is ok on windows).")
            return
        serial_port = open(self.config['port'])
        attrs = termios.tcgetattr(serial_port)
        attrs[2] = attrs[2] & ~termios.HUPCL
        termios.tcsetattr(serial_port, termios.TCSAFLUSH, attrs)
        serial_port.close()

    async def _clear_read_buffer(self):
        """Clear read buffer."""
        # pylint: disable-msg=protected-access
        if self.debug and self._reader._buffer:
            # pylint: disable-msg=protected-access
            self.debug_log("Flushed: %s%s", self._reader._buffer, "".join(" 0x%02x" % b for b in self._reader._buffer))
        if hasattr(self._writer.transport, "_serial"):
            # pylint: disable-msg=protected-access
            self._writer.transport._serial.reset_input_buffer()
        # pylint: disable-msg=protected-access
        self._reader._buffer = bytearray()
        # pylint: disable-msg=protected-access
        self._reader._maybe_resume_transport()

    # pylint: disable-msg=too-many-statements
    # pylint: disable-msg=too-many-branches
    async def initialize(self):
        """Initialise platform."""
        async with self._bus_lock:

            await super().initialize()

            if self.config['connection'] == "serial":
                self.log.info("Connecting to %s at %sbps", self.config['port'], self.config['baud'])
                if self.config['disable_dtr']:
                    self._disable_dts_on_start_of_serial()
                connector = self.machine.clock.open_serial_connection(
                    url=self.config['port'], baudrate=self.config['baud'], limit=0, do_not_open=True)
            else:
                self.log.info("Connecting to %s:%s", self.config['network_host'], self.config['network_port'])
                connector = self.machine.clock.open_connection(self.config['network_host'], self.config['network_port'])

            self._reader, self._writer = await connector

            if self.config['connection'] == "serial":
                if self.config['disable_dtr']:
                    # pylint: disable-msg=protected-access
                    self._writer.transport._serial.dtr = None
                # pylint: disable-msg=protected-access
                self._writer.transport._serial.open()

            # give the serial a few ms to read the first bytes
            await asyncio.sleep(.1)

            while True:
                # reset platform
                self.debug_log("Sending reset.")
                await self._clear_read_buffer()
                # send command
                self.send_byte(LisyDefines.GeneralReset)
                try:
                    return_code = await asyncio.wait_for(self._read_byte(), timeout=0.5)
                except asyncio.TimeoutError:
                    self.warning_log("Reset of LISY failed. Did not get a response in 500ms. Will retry.")
                    continue
                if return_code != 0:
                    # reset failed
                    self.warning_log("Reset of LISY failed. Got %s instead of 0. Will retry.", return_code)
                    continue

                # get type (system 1 vs system 80)
                self.send_byte(LisyDefines.InfoGetConnectedLisyHardware)
                hardware_name = await self._read_string()

                self._coils_start_at_one = bool(hardware_name == b'LISY80')

                self.send_byte(LisyDefines.InfoLisyVersion)
                lisy_version = await self._read_string()
                self.send_byte(LisyDefines.InfoGetApiVersion)
                api_version = await self._read_string()

                self.debug_log("Connected to %s hardware. LISY version: %s. API version: %s.",
                               hardware_name, lisy_version, api_version)

                if not lisy_version:
                    self.error_log("Failed to read lisy_version from LISY. Got %s", lisy_version)
                    continue

                self._lisy_version = lisy_version.decode()

                if api_version:
                    self.api_version = StrictVersion(api_version.decode())
                else:
                    self.error_log("Failed to read api_version from LISY. Got %s", api_version)
                    continue

                # if we made it here reset succeeded
                break

            self._hardware_name = hardware_name.decode()

            self.machine.variables.set_machine_var("lisy_hardware", self._hardware_name)
            '''machine_var: lisy_hardware

            desc: Connected LISY hardware (I.e. LISY1, LISY80 or APC).
            '''
            self.machine.variables.set_machine_var("lisy_version", self._lisy_version)
            '''machine_var: lisy_version

            desc: LISY version.
            '''
            self.machine.variables.set_machine_var("lisy_api_version", api_version)
            '''machine_var: lisy_api_version

            desc: LISY API version.
            '''

            # get number of lamps
            self.send_byte(LisyDefines.InfoGetNumberOfLamps)
            self._number_of_lamps = await self._read_byte()

            # get number of solenoids
            self.send_byte(LisyDefines.InfoGetNumberOfSolenoids)
            self._number_of_solenoids = await self._read_byte()

            # get number of displays
            self.send_byte(LisyDefines.InfoGetNumberOfDisplays)
            self._number_of_displays = await self._read_byte()

            # get number of switches
            self.send_byte(LisyDefines.InfoGetSwitchCount)
            self._number_of_switches = await self._read_byte()

            if self.api_version >= StrictVersion("0.10"):
                # get number of modern lights
                self.send_byte(LisyDefines.GetModernLightsCount)
                # in api version 10+ this returns two bytes
                self._number_of_modern_lights = await self._read_two_bytes()
            elif self.api_version >= StrictVersion("0.9"):
                # get number of modern lights
                self.send_byte(LisyDefines.GetModernLightsCount)
                self._number_of_modern_lights = await self._read_byte()
            else:
                self._number_of_modern_lights = 0

            if self._number_of_modern_lights > 0:
                self._light_system = PlatformBatchLightSystem(self.machine.clock,
                                                              self._send_multiple_light_update,
                                                              self.machine.config['mpf'][
                                                                  'default_light_hw_update_hz'],
                                                              self.config['max_led_batch_size'])
                self._light_system.start()

            self.debug_log("Number of lamps: %s. Number of coils: %s. Numbers of display: %s. Number of switches: %s "
                           "Number of modern lights: %s",
                           self._number_of_lamps, self._number_of_solenoids, self._number_of_displays,
                           self._number_of_switches, self._number_of_modern_lights)

            # initially read all switches
            self.debug_log("Reading all switches.")
            for number in range(self._number_of_switches):
                self.send_byte(LisyDefines.SwitchesGetStatusOfSwitch, bytes([number]))
                state = await self._read_byte()
                if state == 2:
                    self.warning_log("Switch %s does not exist in platform.", number)
                elif state > 2:
                    raise AssertionError("Invalid switch {}. Got response: {}".format(number, state))

                self._inputs[str(number)] = state == 1

            self._watchdog_task = self.machine.clock.loop.create_task(self._watchdog())
            self._watchdog_task.add_done_callback(Util.raise_exceptions)

            self.debug_log("Init of LISY done.")

    async def _send_multiple_light_update(self, sequential_brightness_list):
        common_fade_ms = sequential_brightness_list[0][2]
        if common_fade_ms < 0:
            common_fade_ms = 0
        fade_time = int(common_fade_ms)

        data = bytearray([int(sequential_brightness_list[0][0].number / 256),
                          sequential_brightness_list[0][0].number % 256,
                          int(fade_time / 255), int(fade_time & 0xFF),
                          len(sequential_brightness_list)])
        for _, brightness, _ in sequential_brightness_list:
            data.append(int(255 * brightness))

        self.send_byte(LisyDefines.FadeModernLights, data)

    async def start(self):
        """Start reading switch changes."""
        self._poll_task = self.machine.clock.loop.create_task(self._poll())
        self._poll_task.add_done_callback(Util.raise_exceptions)

    def stop(self):
        """Stop platform."""
        super().stop()
        if self._poll_task:
            self._poll_task.cancel()
            self._poll_task = None

        if self._watchdog_task:
            self._watchdog_task.cancel()
            self._watchdog_task = None

        if self._reader:
            self._writer.close()
            self._reader = None
            self._writer = None

    async def _poll(self):
        sleep_time = 1.0 / self.config['poll_hz']
        while True:
            async with self._bus_lock:
                self.send_byte(LisyDefines.SwitchesGetChangedSwitches)
                status = await self._read_byte()
            if status == 127:
                # no changes. sleep according to poll_hz
                await asyncio.sleep(sleep_time)
            else:
                # bit 7 is state
                switch_state = 1 if status & 0b10000000 else 0
                # bits 0-6 are the switch number
                switch_num = status & 0b01111111

                # tell the switch controller about the new state
                self.machine.switch_controller.process_switch_by_num(str(switch_num), switch_state, self)

                # store in dict as well
                self._inputs[str(switch_num)] = bool(switch_state)

    async def _watchdog(self):
        """Periodically send watchdog."""
        while True:
            # send watchdog
            async with self._bus_lock:
                self.send_byte(LisyDefines.GeneralWatchdog)
                response = await self._read_byte()
                if response != 0:
                    self.warning_log("Watchdog returned %s instead 0", response)
            # sleep 500ms
            await asyncio.sleep(.5)

    # pylint: disable-msg=too-many-arguments
    def _configure_hardware_rule(self, coil: DriverSettings, switch1: SwitchSettings,
                                 switch2: Optional[SwitchSettings], flags1, flags2):
        """Configure hardware rule in LISY."""
        if coil.pulse_settings.duration > 255:
            raise AssertionError("Pulse settings to long for LISY protocol. Got pulse_settings: {}".format(
                coil.pulse_settings))

        coil.hw_driver.has_rule = True
        coil.hw_driver.configure_recycle(coil.pulse_settings.duration * 2 if coil.recycle else
                                         coil.pulse_settings.duration)

        data = bytearray([coil.hw_driver.index,
                          switch1.hw_switch.index + (0x80 if switch1.invert else 0),
                          switch2.hw_switch.index if switch2 else 0 + 0x80 if switch2 and switch2.invert else 0,
                          0,
                          int(coil.pulse_settings.duration),
                          int(coil.pulse_settings.power * 255),
                          int(coil.hold_settings.power * 255) if coil.hold_settings else 0,
                          flags1,
                          flags2,
                          0
                          ])
        self.send_byte(LisyDefines.ConfigureHardwareRuleForSolenoid, data)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and enable and release rule on driver."""
        assert coil.hold_settings.power > 0
        self._configure_hardware_rule(coil, enable_switch, None, 3, 0)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                      eos_switch: SwitchSettings, coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver.

        Pulses a driver when a switch is hit. When the switch is released
        the pulse is canceled and the driver gets disabled. When the eos_switch is hit the pulse is canceled
        and the driver becomes disabled. Typically used on the main coil for dual-wound coil flippers with eos switch.
        """
        assert coil.hold_settings is None
        self._configure_hardware_rule(coil, enable_switch, eos_switch, 3, 2)

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver becomes disabled. When the eos_switch is hit the pulse is canceled
        and the driver becomes enabled (likely with PWM).
        Typically used on the coil for single-wound coil flippers with eos switch.
        """
        self._configure_hardware_rule(coil, enable_switch, eos_switch, 3, 2)

    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and release rule to driver."""
        assert not coil.hold_settings or coil.hold_settings.power == 0
        self._configure_hardware_rule(coil, enable_switch, None, 3, 0)

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit rule on driver."""
        assert not coil.hold_settings or coil.hold_settings.power == 0
        self._configure_hardware_rule(coil, enable_switch, None, 1, 0)

    def clear_hw_rule(self, switch: SwitchSettings, coil: DriverSettings):
        """Clear hw rule for driver."""
        del switch
        if not coil.hw_driver.has_rule:
            return
        coil.hw_driver.has_rule = False

        data = bytearray([coil.hw_driver.index,
                          0,
                          0,
                          0,
                          0,
                          0,
                          0,
                          0,
                          0,
                          0
                          ])
        self.send_byte(LisyDefines.ConfigureHardwareRuleForSolenoid, data)

    def configure_light(self, number: str, subtype: str, config, platform_settings: dict) -> LightPlatformInterface:
        """Configure light on LISY."""
        del platform_settings, config
        assert self._number_of_lamps is not None
        assert self._number_of_modern_lights is not None

        if subtype is None or subtype == "matrix":
            if not self._coils_start_at_one:
                if 0 < int(number) >= self._number_of_lamps:
                    raise AssertionError("LISY only has {} lamps. Cannot configure lamp {} (zero indexed).".
                                         format(self._number_of_lamps, number))
            else:
                if 1 < int(number) > self._number_of_lamps:
                    raise AssertionError("LISY only has {} lamps. Cannot configure lamp {} (one indexed).".
                                         format(self._number_of_lamps, number))

            return LisySimpleLamp(int(number), self)
        if subtype == "light":
            if 0 < int(number) >= self._number_of_modern_lights:
                raise AssertionError("LISY only has {} modern lights. Cannot configure light {}.".
                                     format(self._number_of_modern_lights, number))
            return LisyModernLight(int(number), self, self._light_system)

        raise self.raise_config_error("Invalid subtype {}".format(subtype), 1)

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Return a single light."""
        return [
            {
                "number": number,
            }
        ]

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> SwitchPlatformInterface:
        """Configure a switch."""
        if number not in self._inputs:
            raise AssertionError("Invalid switch number {}. Platform reports the following switches as "
                                 "valid: {}".format(number, list(self._inputs.keys())))

        return LisySwitch(config=config, number=number)

    async def get_hw_switch_states(self):
        """Return current switch states."""
        return self._inputs

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict) -> DriverPlatformInterface:
        """Configure a driver."""
        assert self._number_of_solenoids is not None
        assert self._number_of_lamps is not None

        if 1 < int(number) > self._number_of_solenoids and int(number) < 100:
            raise AssertionError("LISY only has {} drivers. Cannot configure driver {} (zero indexed).".
                                 format(self._number_of_solenoids, number))
        if not self._coils_start_at_one:
            if 100 < int(number) >= self._number_of_lamps + 100:
                raise AssertionError("LISY only has {} lamps. Cannot configure lamp driver {} (zero indexed).".
                                     format(self._number_of_lamps, number))
        else:
            if 101 < int(number) > self._number_of_lamps + 100:
                raise AssertionError("LISY only has {} lamps. Cannot configure lamp driver {} (one indexed).".
                                     format(self._number_of_lamps, number))

        driver = LisyDriver(config=config, number=number, platform=self)
        recycle_time = config.default_pulse_ms * 2 if config.default_recycle else config.default_pulse_ms
        driver.configure_recycle(recycle_time)
        return driver

    async def configure_segment_display(self, number: str, platform_settings) \
            -> SegmentDisplaySoftwareFlashPlatformInterface:
        """Configure a segment display."""
        del platform_settings
        assert self._number_of_displays is not None

        if 0 < int(number) >= self._number_of_displays:
            raise AssertionError("Invalid display number {}. Hardware only supports {} displays (indexed with 0)".
                                 format(number, self._number_of_displays))

        display = LisyDisplay(int(number), self)
        await display.initialize()
        self._handle_software_flash(display)
        return display

    def configure_hardware_sound_system(self) -> HardwareSoundPlatformInterface:
        """Configure hardware sound."""
        return LisySound(self)

    def send_byte(self, cmd: int, byte: bytes = None):
        """Send a command with optional payload."""
        assert self._writer is not None

        if not byte:
            byte = bytes()

        if self._send_length_of_command:
            length = len(byte)  # include command and length byte
            byte = bytes([length]) + byte

        cmd_str = bytes([cmd])
        cmd_str += byte
        self.debug_log("Sending 0x%02x%s (Cmd: %s)", cmd, "".join(" 0x%02x" % b for b in byte), cmd)
        self._writer.write(cmd_str)

    async def send_byte_and_read_response(self, cmd: int, byte: bytes = None, read_bytes=0):
        """Send byte and read response."""
        assert self._reader is not None
        async with self._bus_lock:
            self.send_byte(cmd, byte)
            return await self._reader.readexactly(read_bytes)

    def send_string(self, cmd: int, string: str):
        """Send a command with null terminated string."""
        assert self._writer is not None

        self.debug_log("Sending %s (0x%02x) %s (%s)", cmd, cmd, string, "".join(" 0x%02x" % ord(b) for b in string))
        self._writer.write(bytes([cmd]) + string.encode() + bytes([0]))

    async def _read_byte(self) -> int:
        """Read one byte."""
        assert self._reader is not None

        self.debug_log("Reading one byte")
        data = await self._reader.readexactly(1)
        self.debug_log("Received %s", ord(data))
        return ord(data)

    async def _read_two_bytes(self) -> int:
        """Read two bytes."""
        assert self._reader is not None

        self.debug_log("Reading two bytes")
        data = await self._reader.readexactly(2)
        self.debug_log("Received %s %s", data[0], data[1])
        return (data[0] << 8) + data[1]

    # pylint: disable-msg=inconsistent-return-statements
    async def _readuntil(self, separator, min_chars: int = 0):
        """Read until separator.

        Args:
        ----
            separator: Read until this separator byte.
            min_chars: Minimum message length before separator
        """
        assert self._reader is not None

        # asyncio StreamReader only supports this from python 3.5.2 on
        buffer = b''
        while True:
            char = await self._reader.readexactly(1)
            buffer += char
            if char == separator and len(buffer) > min_chars:
                return buffer

    async def _read_string(self) -> bytes:
        """Read zero terminated string."""
        self.debug_log("Reading zero terminated string")
        data = await self._readuntil(b'\x00')
        # remove terminator
        data = data[:-1]
        self.debug_log("Received %s", data)
        return data

    def get_info_string(self):
        """Dump infos about LISY platform."""
        infos = ""
        if self.config['connection'] == "serial":
            infos += "LISY connected via serial on {} at {}bps\n".format(self.config['port'], self.config['baud'])
        else:
            infos += "LISY connected via network at {}:{}\n".format(self.config['network_host'],
                                                                    self.config['network_port'])
        infos += "Hardware: {} Lisy Version: {} API Version: {}\n".format(self._hardware_name, self._lisy_version,
                                                                          self.api_version)
        infos += "Input count: {} Input map: {}\n".format(self._number_of_switches, sorted(list(self._inputs.keys()),
                                                                                           key=int))
        infos += "Coil count: {}\n".format(self._number_of_solenoids)
        infos += "Modern lights count: {}\n".format(self._number_of_modern_lights)
        infos += "Traditional lights count: {}\n".format(self._number_of_lamps)
        infos += "Display count: {}\n".format(self._number_of_displays)
        return infos



"""Manages operator controllable settings."""
from collections import namedtuple

from typing import Dict, List

from mpf.core.utility_functions import Util

from mpf.core.mpf_controller import MpfController

SettingEntry = namedtuple("SettingEntry", ["name", "label", "sort", "machine_var", "default", "values"])


class SettingsController(MpfController):

    """Manages operator controllable settings."""

    # needed here so the auto-detection of child classes works
    module_name = 'SettingsController'
    config_name = 'settings_controller'

    def __init__(self, machine):
        """Initialise settings controller."""
        super().__init__(machine)

        # start with default settings
        self._settings = {}     # type: Dict[str, SettingEntry]
        """Dictionary of available settings."""

        self._add_entries_from_config()

    def _add_entries_from_config(self):
        # add entries from config
        self.config = self.machine.config.get('settings', {})
        for name, settings in self.config.items():
            settings = self.machine.config_validator.validate_config("settings", settings)
            if not settings['machine_var']:
                settings['machine_var'] = name
            values = {}
            # convert types
            for key, value in settings['values'].items():
                values[Util.convert_to_type(key, settings['key_type'])] = value

            # convert default key
            settings['default'] = Util.convert_to_type(settings['default'], settings['key_type'])

            self.add_setting(SettingEntry(name, settings['label'], settings['sort'], settings['machine_var'],
                                          settings['default'], values))

    def add_setting(self, setting: SettingEntry):
        """Add a setting."""
        self._settings[setting.name] = setting

    def get_settings(self) -> List[SettingEntry]:
        """Return all available settings."""
        sorted_list = list(self._settings.values())
        sorted_list.sort(key=lambda x: x.sort)
        return sorted_list

    def get_setting_value_label(self, setting_name):
        """Return label for value."""
        value = self.get_setting_value(setting_name)
        return self._settings[setting_name].values.get(value, "invalid")

    def __getattr__(self, item):
        """Return setting."""
        if "_settings" not in self.__dict__ or item not in self.__dict__['_settings']:
            raise AttributeError()
        return self.get_setting_value(item)

    def get_setting_machine_var(self, setting_name):
        """Return machine var name."""
        return self._settings[setting_name].machine_var

    def get_setting_value(self, setting_name):
        """Return the current value of a setting."""
        if setting_name not in self._settings:
            raise AssertionError("Invalid setting {}".format(setting_name))

        if not self.machine.variables.is_machine_var(self._settings[setting_name].machine_var):
            value = self._settings[setting_name].default
        else:
            value = self.machine.variables.get_machine_var(self._settings[setting_name].machine_var)

        self.debug_log("Retrieving value: {}={}".format(setting_name, value))

        return value

    def set_setting_value(self, setting_name, value):
        """Set the value of a setting."""
        self.debug_log("New value: {}={}".format(setting_name, value))

        if setting_name not in self._settings:
            raise AssertionError("Invalid setting {}".format(setting_name))

        if value not in self._settings[setting_name].values:
            raise AssertionError("Invalid value {} for setting {}".format(value, setting_name))

        self.machine.variables.configure_machine_var(name=self._settings[setting_name].machine_var, persist=True)
        self.machine.variables.set_machine_var(name=self._settings[setting_name].machine_var, value=value)



"""Test logic blocks."""
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from mpf.tests.MpfTestCase import test_config_directory


class TestLogicBlocks(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/logic_blocks/'

    def test_mode_selection_with_counters(self):
        self.mock_event("qualify_start_mode1")
        self.mock_event("qualify_start_mode2")
        self.start_game()
        self.start_mode("mode3")
        # advance both counters to 2/3
        self.post_event("qualify1_count")
        self.post_event("qualify1_count")
        self.post_event("qualify2_count")
        self.post_event("qualify2_count")

        # post the final even for both of them
        self.machine.switch_controller.process_switch("s_qualify1", 1, True)
        self.machine.switch_controller.process_switch("s_qualify2", 1, True)
        self.advance_time_and_run()
        self.assertEventCalled("qualify_start_mode1")
        self.assertEventNotCalled("qualify_start_mode2")

    @test_config_directory("tests/machine_files/counters/")
    def test_subscription_on_counter_values(self):
        self.start_game()
        self.start_mode("mode1")
        self.assertLightColor("l_chest_matrix_green_2", "black")
        self.assertLightColor("l_chest_matrix_green_3", "black")
        self.assertLightColor("l_chest_matrix_green_4", "black")
        self.assertLightColor("l_chest_matrix_green_5", "black")
        self.post_event("count_up")
        self.advance_time_and_run(.1)
        self.assertLightColor("l_chest_matrix_green_2", "black")
        self.assertLightColor("l_chest_matrix_green_3", "black")
        self.assertLightColor("l_chest_matrix_green_4", "black")
        self.assertLightColor("l_chest_matrix_green_5", "green")
        self.post_event("count_up")
        self.advance_time_and_run(.1)
        self.assertLightColor("l_chest_matrix_green_2", "black")
        self.assertLightColor("l_chest_matrix_green_3", "black")
        self.assertLightColor("l_chest_matrix_green_4", "green")
        self.assertLightColor("l_chest_matrix_green_5", "green")
        self.post_event("count_up")
        self.advance_time_and_run(.1)
        self.assertLightColor("l_chest_matrix_green_2", "black")
        self.assertLightColor("l_chest_matrix_green_3", "green")
        self.assertLightColor("l_chest_matrix_green_4", "green")
        self.assertLightColor("l_chest_matrix_green_5", "green")
        self.post_event("count_up")
        self.advance_time_and_run(.1)
        self.assertLightColor("l_chest_matrix_green_2", "green")
        self.assertLightColor("l_chest_matrix_green_3", "green")
        self.assertLightColor("l_chest_matrix_green_4", "green")
        self.assertLightColor("l_chest_matrix_green_5", "green")

        self.drain_all_balls()
        self.advance_time_and_run()
        self.start_mode("mode1")

        self.assertLightColor("l_chest_matrix_green_2", "black")
        self.assertLightColor("l_chest_matrix_green_3", "black")
        self.assertLightColor("l_chest_matrix_green_4", "black")
        self.assertLightColor("l_chest_matrix_green_5", "black")

    def test_counter_with_lights(self):
        self.start_game()
        self.post_event("start_mode2")
        self.advance_time_and_run()

        self.assertLightColor("led1", "white")
        self.assertLightColor("led2", "black")
        self.assertLightColor("led3", "black")

        # nothing happens because it is disabled
        self.post_event("counter_with_lights_count")
        self.advance_time_and_run()
        self.assertLightColor("led1", "white")
        self.assertLightColor("led2", "black")
        self.assertLightColor("led3", "black")

        # advance
        self.post_event("counter_with_lights_enable")
        self.post_event("counter_with_lights_count")
        self.advance_time_and_run()
        self.assertLightColor("led1", "black")
        self.assertLightColor("led2", "white")
        self.assertLightColor("led3", "black")

        # stop mode
        self.post_event("stop_mode2")
        self.advance_time_and_run()

        # all off
        self.assertLightColor("led1", "black")
        self.assertLightColor("led2", "black")
        self.assertLightColor("led3", "black")

        # restart mode. should restore state
        self.post_event("start_mode2")
        self.advance_time_and_run()
        self.assertLightColor("led1", "black")
        self.assertLightColor("led2", "white")
        self.assertLightColor("led3", "black")

        # and complete
        self.post_event("counter_with_lights_count")
        self.advance_time_and_run()
        self.assertLightColor("led1", "black")
        self.assertLightColor("led2", "black")
        self.assertLightColor("led3", "white")

    def test_accrual_random_advance(self):
        self.start_game()
        self.mock_event("accrual1_complete1")

        # should do nothing
        self.post_event("accrual1_random_advance")
        self.assertEqual([False, False, False], self.machine.accruals["accrual1"].value)

        # enable accrual
        self.post_event("accrual1_enable")

        # complete one step
        self.post_event("accrual1_step1a")
        self.assertEqual([True, False, False], self.machine.accruals["accrual1"].value)

        # should advance one of the remaining steps
        self.post_event("accrual1_random_advance")

        # exactly two steps should be hit
        self.assertEqual(2, sum(self.machine.accruals["accrual1"].value))
        self.assertEventNotCalled("accrual1_complete1")

        # should complete the accrual
        self.post_event("accrual1_random_advance")

        self.assertEventCalled("accrual1_complete1")

    def test_accruals_simple(self):
        self.start_game()
        self.mock_event("accrual1_complete1")
        self.mock_event("accrual1_hit")
        self.mock_event("accrual1_complete2")

        # accrual should not yet work
        self.post_event("accrual1_step1a")
        self.post_event("accrual1_step2b")
        self.post_event("accrual1_step3c")

        self.assertEqual(0, self._events["accrual1_complete1"])
        self.assertEqual(0, self._events["accrual1_complete2"])

        # enable accrual
        self.post_event("accrual1_enable")

        # step2
        self.post_event("accrual1_step2a")
        self.assertEqual(0, self._events["accrual1_complete1"])
        self.assertEqual(1, self._events["accrual1_hit"])

        # step1
        self.post_event("accrual1_step1c")
        self.post_event("accrual1_step1b")
        self.assertEqual(0, self._events["accrual1_complete1"])
        self.assertEqual(2, self._events["accrual1_hit"])

        # step 3
        self.post_event("accrual1_step3c")

        # accrual should fire
        self.assertEqual(1, self._events["accrual1_complete1"])
        self.assertEqual(1, self._events["accrual1_complete2"])

        # should not work again
        self.post_event("accrual1_step1a")
        self.post_event("accrual1_step2a")
        self.post_event("accrual1_step3a")
        self.assertEqual(1, self._events["accrual1_complete1"])
        self.assertEqual(1, self._events["accrual1_complete2"])

        # reset but do not enable yet
        self.post_event("accrual1_reset")

        # nothing should happen
        self.post_event("accrual1_step1a")
        self.post_event("accrual1_step2a")
        self.post_event("accrual1_step3a")
        self.assertEqual(1, self._events["accrual1_complete1"])
        self.assertEqual(1, self._events["accrual1_complete2"])

        # enable for one step
        self.post_event("accrual1_enable")
        self.post_event("accrual1_step1a")

        # disable for next
        self.post_event("accrual1_disable")
        self.post_event("accrual1_step2a")

        # enable for third step
        self.post_event("accrual1_enable")
        self.post_event("accrual1_step3a")

        # should not complete yet
        self.assertEqual(1, self._events["accrual1_complete1"])
        self.assertEqual(1, self._events["accrual1_complete2"])

        self.post_event("accrual1_step2a")

        # but now
        self.assertEqual(2, self._events["accrual1_complete1"])
        self.assertEqual(2, self._events["accrual1_complete2"])

    def test_counter_simple_down(self):
        self.start_game()
        self.mock_event("logicblock_counter1_complete")
        self.mock_event("logicblock_counter1_hit")

        self.post_event("counter1_enable")
        for i in range(4):
            self.post_event("counter1_count")
            self.assertEqual(0, self._events["logicblock_counter1_complete"])

        # nothing should happen when disabled
        self.post_event("counter1_disable")
        for i in range(10):
            self.post_event("counter1_count")
            self.assertEqual(0, self._events["logicblock_counter1_complete"])
        self.post_event("counter1_enable")

        self.post_event("counter1_count")
        self.assertEqual(1, self._events["logicblock_counter1_complete"])
        self.assertEqual(5, self._events["logicblock_counter1_hit"])

        # it should disable
        self.post_event("counter1_count")
        self.assertEqual(1, self._events["logicblock_counter1_complete"])
        self.assertEqual(5, self._events["logicblock_counter1_hit"])

        self.post_event("counter1_restart")

        for i in range(4):
            self.post_event("counter1_count")

        # 4 more hits but not completed
        self.assertEqual(1, self._events["logicblock_counter1_complete"])
        self.assertEqual(9, self._events["logicblock_counter1_hit"])

        # reset
        self.post_event("counter1_reset")
        for i in range(4):
            self.post_event("counter1_count")

        # another 4 hits still not complete
        self.assertEqual(1, self._events["logicblock_counter1_complete"])
        self.assertEqual(13, self._events["logicblock_counter1_hit"])

        # and complete again
        self.post_event("counter1_count")
        self.assertEqual(2, self._events["logicblock_counter1_complete"])
        self.assertEqual(14, self._events["logicblock_counter1_hit"])

    def test_sequence_simple(self):
        self.start_game()
        self.mock_event("sequence1_complete")
        self.mock_event("logicblock_sequence1_hit")

        self.post_event("sequence1_enable")

        # wrong order
        self.post_event("sequence1_step3a")
        self.post_event("sequence1_step2a")
        self.post_event("sequence1_step1b")
        self.assertEqual(0, self._events["sequence1_complete"])
        self.assertEqual(1, self._events["logicblock_sequence1_hit"])

        # still not
        self.post_event("sequence1_step3b")
        self.post_event("sequence1_step1a")
        self.assertEqual(0, self._events["sequence1_complete"])
        self.assertEqual(1, self._events["logicblock_sequence1_hit"])

        # only 1 so far. now step2
        self.post_event("sequence1_step2a")
        self.assertEqual(0, self._events["sequence1_complete"])
        self.assertEqual(2, self._events["logicblock_sequence1_hit"])

        # and step 3
        self.post_event("sequence1_step3b")
        self.assertEqual(1, self._events["sequence1_complete"])
        self.assertEqual(3, self._events["logicblock_sequence1_hit"])

        # should be disabled
        self.post_event("sequence1_step1a")
        self.post_event("sequence1_step2a")
        self.post_event("sequence1_step3a")
        self.assertEqual(1, self._events["sequence1_complete"])

        # enable and reset
        self.post_event("sequence1_enable")
        self.post_event("sequence1_reset")

        # reset inbetween
        self.post_event("sequence1_step1a")
        self.post_event("sequence1_step2a")
        self.post_event("sequence1_reset")
        self.post_event("sequence1_step3a")

        # nothing
        self.assertEqual(1, self._events["sequence1_complete"])

        # again
        self.post_event("sequence1_step1a")
        self.assertEqual(1, self._events["sequence1_complete"])
        self.post_event("sequence1_step2a")
        self.assertEqual(1, self._events["sequence1_complete"])
        self.post_event("sequence1_step3a")
        self.assertEqual(2, self._events["sequence1_complete"])

    def test_counter_in_mode(self):
        self.start_game()
        self.mock_event("counter2_complete")
        self.mock_event("counter2_hit")

        for i in range(10):
            self.post_event("counter2_count")
            self.assertEqual(0, self._events["counter2_complete"])

        self.post_event("start_mode1")
        self.assertTrue("mode1" in self.machine.modes)

        for i in range(2):
            self.post_event("counter2_count")
            self.assertEqual(i + 1, self._events["counter2_hit"])
            self.assertEqual(0, self._events["counter2_complete"])
            self.assertEventCalledWith("counter2_hit", count=i + 1, remaining=2 - i, hits=i+1)

        self.post_event("counter2_count")
        self.assertEqual(1, self._events["counter2_complete"])
        self.assertEventCalledWith("counter2_hit", count=3, hits=3, remaining=0)

        # should run again
        for i in range(2):
            self.post_event("counter2_count")
            self.assertEqual(i + 4, self._events["counter2_hit"])
            self.assertEqual(1, self._events["counter2_complete"])

        self.post_event("counter2_count")
        self.assertEqual(2, self._events["counter2_complete"])

        # stop mode
        self.post_event("stop_mode1")

        # nothing should happen any more
        for i in range(10):
            self.post_event("counter2_count")
            self.assertEqual(2, self._events["counter2_complete"])
            self.assertEqual(6, self._events["counter2_hit"])

    def test_counter_control_events(self):
        '''
        Tests the add, subtract, and set_value control events
        for the Counter class.
        '''
        def reset_event_mocks():
            # Reset mocks
            self.mock_event("counter6_complete")
            self.mock_event("counter6_hit")
            self.mock_event("counter7_complete")
            self.mock_event("counter7_hit")
        self.start_game()
        reset_event_mocks()
        # Start mode with control events and counter6
        self.post_event("start_mode4")
        self.assertTrue("mode4" in self.machine.modes)

        # Adds zero to the counter 10 times, counter should not reach completion
        for i in range(10):
            self.post_event("increase_counter6_0")
            self.assertEqual(0, self._events["counter6_complete"])
        # Counts the counter once, and then adds 3 to it 3 times,
        # The last adding of three should cause the counter to complete once
        for i in range(0, 2):
            self.post_event("increase_counter6_3")
            self.assertEqual(0, self._events["counter6_complete"])
        self.post_event("counter6_count")
        self.assertEventCalledWith("counter6_hit", count=7, hits=7, remaining=3)
        self.post_event("increase_counter6_3")
        self.assertEqual(1, self._events["counter6_complete"])

        # Test the adding of five to the counter
        reset_event_mocks()
        self.post_event("increase_counter6_5")
        self.assertEqual(0, self._events["counter6_complete"])
        self.post_event("increase_counter6_5")
        self.assertEqual(1, self._events["counter6_complete"])

        # Test subtraction
        reset_event_mocks()
        self.post_event("counter6_count")
        self.assertEventCalledWith("counter6_hit", count=1, hits=1, remaining=9)
        self.post_event("reduce_counter6_5")
        self.post_event("counter6_count")
        self.assertEventCalledWith("counter6_hit", count=-3, hits=-3, remaining=13)
        self.post_event("reduce_counter6_3")
        self.post_event("counter6_count")
        self.assertEventCalledWith("counter6_hit", count=-5, hits=-5, remaining=15)
        self.post_event("reduce_counter6_0")
        self.post_event("counter6_count")
        self.assertEventCalledWith("counter6_hit", count=-4, hits=-4, remaining=14)

        # Test Setting the Counter to a value
        reset_event_mocks()
        # Make sure that the counter holds a nonzero value
        self.post_event("counter6_count")
        self.assertEventCalledWith("counter6_hit", count=-3, hits=-3, remaining=13)
        self.post_event("set_counter6_0")
        self.post_event("counter6_count")
        self.assertEventCalledWith("counter6_hit", count=1, hits=1, remaining=9)
        # Set the counter to a value above the completion value
        self.assertEqual(0, self._events["counter6_complete"])
        self.post_event("set_counter6_25")
        self.assertEqual(1, self._events["counter6_complete"])

        # Test using counter with direction down
        # Test increasing and reducing
        reset_event_mocks()
        self.post_event("counter7_count")
        self.assertEventCalledWith("counter7_hit", count=4, hits=1, remaining=4)
        self.post_event("increase_counter7_5")
        self.post_event("counter7_count")
        self.assertEventCalledWith("counter7_hit", count=8, hits=-3, remaining=8)
        self.post_event("reduce_counter7_5")
        self.post_event("counter7_count")
        self.assertEventCalledWith("counter7_hit", count=2, hits=3, remaining=2)
        self.assertEqual(0, self._events["counter7_complete"])
        self.post_event("reduce_counter7_3")
        self.assertEqual(1, self._events["counter7_complete"])

        # Test setting the value with direction down counter
        reset_event_mocks()
        self.assertEqual(0, self._events["counter7_complete"])
        self.post_event("set_counter7_negative25")
        self.assertEqual(1, self._events["counter7_complete"])
        self.post_event("set_counter7_0")
        self.assertEqual(2, self._events["counter7_complete"])
        self.post_event("set_counter7_3")
        self.post_event("counter7_count")
        self.assertEventCalledWith("counter7_hit", count=2, hits=3, remaining=2)

        self.assertPlaceholderEvaluates(2, "device.counters.counter7.value")

        # nothing happens because machine.test2 is undefined
        self.post_event("set_counter_placeholder")
        self.assertPlaceholderEvaluates(2, "device.counters.counter7.value")

        self.machine.variables.set_machine_var("test2", 4)
        self.post_event("set_counter_placeholder")
        self.assertPlaceholderEvaluates(4, "device.counters.counter7.value")

        self.post_event("subtract_counter_placeholder")
        self.assertPlaceholderEvaluates(4, "device.counters.counter7.value")

        self.machine.variables.set_machine_var("test3", 3)
        self.post_event("subtract_counter_placeholder")
        self.assertPlaceholderEvaluates(1, "device.counters.counter7.value")

        self.post_event("add_counter_placeholder")
        self.assertPlaceholderEvaluates(1, "device.counters.counter7.value")

        self.machine.variables.set_machine_var("test4", 1)
        self.post_event("add_counter_placeholder")
        self.assertPlaceholderEvaluates(2, "device.counters.counter7.value")

    def test_logic_block_outside_game(self):
        self.mock_event("logicblock_accrual2_complete")

        # should work before game
        self.post_event("accrual2_step1")
        self.post_event("accrual2_step2")
        self.assertEqual(1, self._events["logicblock_accrual2_complete"])
        self.post_event("accrual2_restart")

        self.start_game()
        # should work during game
        self.post_event("accrual2_step1")
        self.post_event("accrual2_step2")
        self.assertEqual(2, self._events["logicblock_accrual2_complete"])
        self.post_event("accrual2_restart")

        self.stop_game()

        # should work after game
        self.post_event("accrual2_step1")
        self.post_event("accrual2_step2")
        self.assertEqual(3, self._events["logicblock_accrual2_complete"])

    def test_no_reset_on_complete(self):
        self.mock_event("logicblock_accrual3_complete")

        # start game
        self.start_game()
        # and enable
        self.post_event("accrual3_enable")

        # should work once
        self.post_event("accrual3_step1")
        self.post_event("accrual3_step2")
        self.assertEqual(1, self._events["logicblock_accrual3_complete"])

        # but not a second time because it disabled
        self.post_event("accrual3_step1")
        self.post_event("accrual3_step2")
        self.assertEqual(1, self._events["logicblock_accrual3_complete"])

        # enable again
        self.post_event("accrual3_enable")

        # still completed
        self.post_event("accrual3_step1")
        self.post_event("accrual3_step2")
        self.assertEqual(1, self._events["logicblock_accrual3_complete"])

        # should work after reset
        self.post_event("accrual3_reset")
        self.post_event("accrual3_step1")
        self.post_event("accrual3_step2")
        self.assertEqual(2, self._events["logicblock_accrual3_complete"])

        # disabled again
        self.post_event("accrual3_reset")
        self.post_event("accrual3_step1")
        self.post_event("accrual3_step2")
        self.assertEqual(2, self._events["logicblock_accrual3_complete"])

        # works after enable
        self.post_event("accrual3_enable")
        self.post_event("accrual3_step1")
        self.post_event("accrual3_step2")
        self.assertEqual(3, self._events["logicblock_accrual3_complete"])

    def test_no_reset_and_no_disable_on_complete(self):
        self.mock_event("logicblock_accrual4_complete")

        # start game
        self.start_game()
        # and enable
        self.post_event("accrual4_enable")
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual4.enabled")
        self.assertPlaceholderEvaluates(False, "device.accruals.accrual4.completed")

        # should work once
        self.post_event("accrual4_step1")
        self.post_event("accrual4_step2")
        self.assertEqual(1, self._events["logicblock_accrual4_complete"])
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual4.enabled")
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual4.completed")

        # enabled but still completed
        self.post_event("accrual4_step1")
        self.post_event("accrual4_step2")
        self.assertEqual(1, self._events["logicblock_accrual4_complete"])
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual4.enabled")
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual4.completed")

        # should work after reset
        self.post_event("accrual4_reset")
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual4.enabled")
        self.assertPlaceholderEvaluates(False, "device.accruals.accrual4.completed")
        self.post_event("accrual4_step1")
        self.post_event("accrual4_step2")
        self.assertEqual(2, self._events["logicblock_accrual4_complete"])
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual4.enabled")
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual4.completed")

    def test_reset_and_no_disable_on_complete(self):
        self.mock_event("logicblock_accrual10_complete")

        # start game
        self.start_game()
        # and enable
        self.post_event("accrual10_enable")
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual10.enabled")
        self.assertPlaceholderEvaluates(False, "device.accruals.accrual10.completed")

        # should work once
        self.post_event("accrual10_step1")
        self.post_event("accrual10_step2")
        self.assertEqual(1, self._events["logicblock_accrual10_complete"])

        # and instantly reset and work again
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual10.enabled")
        self.assertPlaceholderEvaluates(False, "device.accruals.accrual10.completed")
        self.post_event("accrual10_step1")
        self.post_event("accrual10_step2")
        self.assertEqual(2, self._events["logicblock_accrual10_complete"])
        self.assertPlaceholderEvaluates(True, "device.accruals.accrual10.enabled")
        self.assertPlaceholderEvaluates(False, "device.accruals.accrual10.completed")

    def test_player_change(self):
        self.mock_event("logicblock_accrual5_complete")

        self.machine.config['game']['balls_per_game'] = self.machine.placeholder_manager.build_int_template(2)

        self.start_two_player_game()
        self.advance_time_and_run()
        self.post_event("start_mode1")
        self.advance_time_and_run(.1)

        # should work during game - player1
        self.assertEqual(1, self.machine.game.player.number)
        self.post_event("accrual5_step1")
        self.post_event("accrual5_step2")
        self.assertEqual(1, self._events["logicblock_accrual5_complete"])

        # player2
        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.post_event("start_mode1")
        self.advance_time_and_run(.1)

        # not yet complete
        self.post_event("accrual5_step1")
        self.assertEqual(1, self._events["logicblock_accrual5_complete"])

        # player1 again
        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.post_event("start_mode1")
        self.advance_time_and_run(.1)

        # nothing should happen because its disabled and completed for player1
        self.post_event("accrual5_step1")
        self.post_event("accrual5_step2")
        self.assertEqual(1, self._events["logicblock_accrual5_complete"])

        # player2 again
        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.post_event("start_mode1")
        self.advance_time_and_run(.1)

        # complete it
        self.post_event("accrual5_step2")
        self.assertEqual(2, self._events["logicblock_accrual5_complete"])

        self.post_event("stop_mode1")
        self.stop_game()

        # does not work after game
        self.post_event("accrual5_step1")
        self.post_event("accrual5_step2")
        self.assertEqual(2, self._events["logicblock_accrual5_complete"])

    def test_counter_hit_window(self):
        self.start_game()
        self.mock_event("logicblock_counter3_complete")
        self.mock_event("counter_counter3_hit")

        self.post_event("counter3_enable")
        for i in range(10):
            self.post_event("counter3_count")
            self.assertEqual(0, self._events["logicblock_counter3_complete"])

        # inside same window. only one hit
        self.assertEqual(1, self._events["counter_counter3_hit"])
        self.assertEqual(0, self._events["logicblock_counter3_complete"])
        self.advance_time_and_run(1)

        for i in range(3):
            self.post_event("counter3_count")
            self.assertEqual(0, self._events["logicblock_counter3_complete"])
            self.assertEqual(2 + i, self._events["counter_counter3_hit"])
            self.advance_time_and_run(1)

        # it should complete
        self.post_event("counter3_count")
        self.assertEqual(1, self._events["logicblock_counter3_complete"])
        self.assertEqual(5, self._events["counter_counter3_hit"])

    def test_counter_template(self):
        self.start_game()
        self.mock_event("logicblock_counter4_complete")
        self.mock_event("counter_counter4_hit")

        self.machine.game.player.hits = 2

        self.post_event("counter4_enable")
        for i in range(2):
            self.assertEqual(0, self._events["logicblock_counter4_complete"])
            self.post_event("counter4_count")

        self.assertEqual(2, self._events["counter_counter4_hit"])
        self.assertEqual(1, self._events["logicblock_counter4_complete"])
        self.advance_time_and_run(1)

        self.machine.variables.set_machine_var("start", 1)
        self.machine.game.player.hits = 5
        self.mock_event("logicblock_counter4_complete")
        self.mock_event("counter_counter4_hit")

        self.post_event("counter4_reset")
        self.post_event("counter4_enable")
        for i in range(4):
            self.assertEqual(0, self._events["logicblock_counter4_complete"])
            self.post_event("counter4_count")

        # inside same window. only one hit
        self.assertEqual(4, self._events["counter_counter4_hit"])
        self.assertEqual(1, self._events["logicblock_counter4_complete"])
        self.advance_time_and_run(1)

    def test_counter_persist(self):
        self.mock_event("logicblock_counter_persist_complete")
        self.mock_event("counter_counter_persist_hit")
        self.start_two_player_game()
        self.post_event("start_mode1")
        self.assertTrue("mode1" in self.machine.modes)
        self.post_event("counter_persist_enable")

        for i in range(3):
            self.post_event("counter_persist_count")
            self.assertEqual(i + 1, self._events["counter_counter_persist_hit"])

        self.assertEqual(0, self._events["logicblock_counter_persist_complete"])

        self.drain_all_balls()
        self.assertPlayerNumber(2)

        for i in range(10):
            self.post_event("counter_persist_count")

        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.post_event("start_mode1")
        self.post_event("counter_persist_enable")

        self.assertEqual(0, self._events["logicblock_counter_persist_complete"])

        for i in range(2):
            self.post_event("counter_persist_count")
            self.assertEqual(i + 4, self._events["counter_counter_persist_hit"])

        self.assertEqual(1, self._events["logicblock_counter_persist_complete"])

    def test_count_without_end(self):
        self.start_game()
        self.post_event("counter5_count")
        self.post_event("counter5_count")
        self.post_event("counter5_count")

        self.assertEqual(3, self.machine.counters["counter5"].value)



"""Contains the DeviceManager base class."""
import asyncio
from collections import OrderedDict

from typing import Callable, Tuple, List, Generator

from mpf.core.utility_functions import Util
from mpf.core.mpf_controller import MpfController

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.device import Device  # pylint: disable-msg=cyclic-import,unused-import


class DeviceManager(MpfController):

    """Manages all the devices in MPF."""

    config_name = "device_manager"

    __slots__ = ["_monitorable_devices", "collections", "device_classes"]

    def __init__(self, machine):
        """Initialize device manager."""
        super().__init__(machine)

        self._monitorable_devices = {}

        self.collections = OrderedDict()
        self.device_classes = OrderedDict()  # collection_name: device_class

        # this has to happen after mode load
        self.machine.events.add_async_handler('init_phase_1',
                                              self._load_device_modules, priority=5)

        self.machine.events.add_handler('init_phase_2',
                                        self.create_machinewide_device_control_events,
                                        priority=2)

    def get_monitorable_devices(self):
        """Return all devices which are registered as monitorable."""
        return self._monitorable_devices

    def register_monitorable_device(self, device):
        """Register a monitorable device.

        Args:
        ----
            device: The device to register.
        """
        if device.collection not in self._monitorable_devices:
            self._monitorable_devices[device.collection] = {}
        self._monitorable_devices[device.collection][device.name] = device

    def notify_device_changes(self, device, notify, old, value):
        """Notify subscribers about changes in a registered device.

        Args:
        ----
            device: The device that changed.
            notify: Attribute name which changed.
            old: The old value.
            value: The new value.

        """
        self.machine.bcp.interface.notify_device_changes(device, notify, old, value)

    async def _load_device_modules(self, **kwargs):
        del kwargs
        # step 1: create devices in machine collection
        self.debug_log("Creating devices...")
        for device_type in self.machine.config['mpf']['device_modules']:
            device_cls = Util.string_to_class(device_type)      # type: Device

            collection_name, config = device_cls.get_config_info()

            self.device_classes[collection_name] = device_cls

            # create the collection
            collection = DeviceCollection(self.machine, collection_name,
                                          device_cls.config_section)

            self.collections[collection_name] = collection
            setattr(self.machine, collection_name, collection)

            # Get the config section for these devices
            config = self.machine.config.get(config, None)

            # create the devices
            if config:
                self.create_devices(collection_name, config)

            # create the default control events
            try:
                self._create_default_control_events(collection)
            except KeyError:
                pass

        self.machine.mode_controller.create_mode_devices()

        # step 2: load config and validate devices
        self.load_devices_config(validate=True)
        await self.machine.mode_controller.load_mode_devices()

        # step 3: initialise devices (mode devices will be initialised when mode is started)
        await self.initialize_devices()

    def stop_devices(self):
        """Stop all devices in the machine."""
        for device_type in self.machine.config['mpf']['device_modules']:
            device_cls = Util.string_to_class(device_type)
            collection_name, _ = device_cls.get_config_info()
            if not hasattr(self.machine, collection_name):
                continue
            for device in getattr(self.machine, collection_name).values():
                if hasattr(device, "stop_device"):
                    device.stop_device()

    def create_devices(self, collection_name, config):
        """Create devices for a collection."""
        cls = self.device_classes[collection_name]

        collection = getattr(self.machine, collection_name)

        # if this device class has a device_class_init classmethod, run it now
        if hasattr(cls, 'device_class_init'):
            # don't want to use try here in case the called method has an error
            cls.device_class_init(self.machine)

        # create the devices
        for device_name in config:

            if not config[device_name] and not cls.allow_empty_configs:
                self.raise_config_error("Device {}:'{}' has an empty config.".format(collection_name, device_name), 2,
                                        context=collection_name + "." + device_name)

            elif not isinstance(config[device_name], dict):
                self.raise_config_error("Device {}:'{}' does not have a valid config. Expected a dictionary.".format(
                    collection_name, device_name), 3, context=collection_name + "." + device_name)

            collection[device_name] = cls(self.machine, device_name)

    def load_devices_config(self, validate=True):
        """Load all devices."""
        if validate:
            for device_type in self.machine.config['mpf']['device_modules']:

                device_cls = Util.string_to_class(device_type)

                collection_name, config_name = device_cls.get_config_info()

                if config_name not in self.machine.config:
                    continue

                # Get the config section for these devices
                collection = getattr(self.machine, collection_name)
                config = self.machine.config[config_name]
                if not config:
                    self.machine.config[config_name] = config = {}
                if not isinstance(config, dict):
                    self.raise_config_error("Format of collection {} is invalid.".format(collection_name), 1)

                # validate config
                for device_name in config:
                    config[device_name] = collection[device_name].prepare_config(config[device_name], False)
                    config[device_name] = collection[device_name].validate_and_parse_config(config[device_name], False)

        for device_type in self.machine.config['mpf']['device_modules']:

            device_cls = Util.string_to_class(device_type)

            collection_name, config_name = device_cls.get_config_info()

            if config_name not in self.machine.config:
                continue

            # Get the config section for these devices
            collection = getattr(self.machine, collection_name)
            config = self.machine.config[config_name]

            # load config
            for device_name in config:
                collection[device_name].load_config(config[device_name])

    async def initialize_devices(self):
        """Initialise devices."""
        futures = []
        for device_type in self.machine.config['mpf']['device_modules']:

            device_cls = Util.string_to_class(device_type)

            collection_name, config_name = device_cls.get_config_info()

            if config_name not in self.machine.config:
                continue

            # Get the config section for these devices
            collection = getattr(self.machine, collection_name)
            config = self.machine.config[config_name]

            # add machine wide
            for device_name in config:
                futures.append(collection[device_name].device_added_system_wide())

        await asyncio.wait(futures)

    # pylint: disable-msg=too-many-nested-blocks
    def get_device_control_events(self, config) -> Generator[Tuple[str, Callable, int, "Device"], None, None]:
        """Scan a config dictionary for control_events.

         Yields events, methods, delays, and devices for all the devices and
         control_events in that config.

        Args:
        ----
            config: An MPF config dictionary (either machine-wide or mode-
                specific).

        Returns a generator of 4-item tuples
        ------------------------------------
            * The event name
            * The callback method of the device
            * The delay in ms
            * The device object
        """
        config_spec = self.machine.config_validator.get_config_spec()
        for collection in self.collections:
            config_section = self.collections[collection].config_section
            if config_section in config:
                for device, settings in iter(config[config_section].items()):

                    control_events = [x for x in settings if
                                      x.endswith('_events') and x != "control_events"]
                    device_obj = self.collections[collection][device]

                    for control_event in control_events:
                        # get events from this device's config
                        if not isinstance(settings[control_event], dict):
                            if config_spec[config_section][control_event] == "ignore":
                                continue
                            raise AssertionError(
                                "Type of {}:{} should be event_handler|event_handler:ms| in config_spec".format(
                                    collection, control_event))
                        # try the new style first
                        try:
                            method = getattr(device_obj, "event_{}".format(control_event[:-7]))
                        except AttributeError:
                            raise AssertionError("Class {} needs to have method event_{} to handle {}".format(
                                device_obj, control_event[:-7], control_event
                            ))
                        for event, delay in settings[control_event].items():
                            yield (event,
                                   method,
                                   delay,
                                   device_obj)

    def create_machinewide_device_control_events(self, **kwargs):
        """Create machine wide control events."""
        del kwargs
        for event, method, delay, _ in (
                self.get_device_control_events(self.machine.config)):

            if delay:
                self.machine.events.add_handler(
                    event=event,
                    handler=self._control_event_handler,
                    callback=method,
                    ms_delay=delay,
                    delay_mgr=self.machine.delay)
            else:
                self.machine.events.add_handler(
                    event=event,
                    handler=method)

    def _control_event_handler(self, callback, ms_delay, delay_mgr=None, **kwargs):
        del kwargs

        self.debug_log("_control_event_handler: callback: %s,", callback)
        delay_mgr.add(ms=ms_delay, callback=callback)

    def _create_default_control_events(self, device_list):
        for device in device_list.values():

            event_prefix = device.class_label + '_' + device.name + '_'
            event_prefix2 = device.collection + '_'

            for method in (self.machine.config['mpf']['device_events']
                           [device.config_section]):
                self.machine.events.add_handler(event=event_prefix + method,
                                                handler=getattr(device,
                                                                method))
                self.machine.events.add_handler(event=event_prefix2 + method,
                                                handler=getattr(device,
                                                                method))


class DeviceCollection(dict):

    """A collection of Devices.

    One instance of this class will be created for each different type of
    hardware device (such as coils, lights, switches, ball devices, etc.).
    """

    __slots__ = ["machine", "name", "config_section", "_tag_cache"]

    def __init__(self, machine, collection, config_section):
        """Initialise device collection."""
        super().__init__()

        self.machine = machine
        self.name = collection
        self.config_section = config_section
        self._tag_cache = dict()

    def __hash__(self):
        """Hash collection."""
        return hash((self.name, self.machine))

    def __delitem__(self, key):
        """Delete item for key."""
        # clear the tag cache
        self._tag_cache = dict()
        return super().__delitem__(key)

    def __getattr__(self, attr):
        """Return device by key.

        This method is DEPRECATED and will be removed soon.
        """
        # We used this to allow the programmer to access a hardware item like
        # self.coils.coilname

        try:
            return self[attr]
        except KeyError:
            raise KeyError('Error: No device exists with the name:', attr)

    def __iter__(self):
        """Iterate collection.

        This method is DEPRECATED and will be removed soon. Use .values() instead.
        """
        for item in self.values():
            yield item

    def items_tagged(self, tag) -> List["Device"]:
        """Return of list of device objects which have a certain tag.

        Args:
        ----
            tag: A string of the tag name which specifies what devices are
                returned.

        Returns a list of device objects. If no devices are found with that tag, it
        will return an empty list.
        """
        items_in_tag_cache = self._tag_cache.get(tag, None)
        if items_in_tag_cache is not None:
            return items_in_tag_cache

        items = [item for item in self.values() if hasattr(item, "tags") and tag in item.tags]
        self._tag_cache[tag] = items
        return items
